// file      : xsd/cxx/parser/name-processor.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2006-2007 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/parser/name-processor.hxx>

#include <cxx/parser/elements.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <cult/containers/set.hxx>

//#include <iostream>

namespace CXX
{
  namespace Parser
  {
    namespace
    {
      //
      //
      typedef Cult::Containers::Set<String> NameSet;

      class Context
      {
      public:
        Context (SemanticGraph::Schema& schema_,
                 SemanticGraph::Path const& file)
            : schema_path_ (file),
              schema (schema_),
              schema_path (schema_path_),
              ns_ (0),
              ns (ns_),
              global_type_names (global_type_names_)
        {
        }

      protected:
        Context (Context& c)
            : schema (c.schema),
              schema_path (c.schema_path),
              ns (c.ns),
              global_type_names (c.global_type_names)
        {
        }

      private:
        SemanticGraph::Path const schema_path_;
        SemanticGraph::Namespace* ns_;

        Cult::Containers::Map<String, NameSet> global_type_names_;

      public:
        SemanticGraph::Schema& schema;
        SemanticGraph::Path const& schema_path;
        SemanticGraph::Namespace*& ns;

        Cult::Containers::Map<String, NameSet>& global_type_names;
      };


      String
      find_name (SemanticGraph::Nameable& n, NameSet& set)
      {
        String base_name (Parser::Context::escape (n.name ()));
        String name (base_name);

        for (UnsignedLong i (1); set.find (name) != set.end (); ++i)
        {
          std::wostringstream os;
          os << i;
          name = base_name + os.str ();
        }

        set.insert (name);
        n.context ().set ("name", name);

        return name;
      }


      //
      //
      struct Member: Traversal::Member
      {
        Member (NameSet& set)
            : set_ (set)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (Parser::Context::skip (m))
            return;

          find_name (m, set_);
        }

      private:
        NameSet& set_;
      };


      //
      //
      struct MemberInRestriction: Traversal::Member
      {
        MemberInRestriction (NameSet& set, SemanticGraph::Complex& base)
            : set_ (set), base_ (base)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (Parser::Context::skip (m))
            return;

          try
          {
            // Try to find corresponding member in one of our bases.
            // This may fail if we use an element that substitutes
            // one in our base.
            //
            Type& bm (find_member (base_, m));
            m.context ().set ("name", bm.context ().get<String> ("name"));
          }
          catch (NotFound const&)
          {
            // Fall back to the standard name assignment.
            //
            find_name (m, set_);
          }
        }

      private:
        struct NotFound {};

        Type&
        find_member (SemanticGraph::Complex& c, Type& m)
        {
          using SemanticGraph::Complex;

          Complex::NamesIteratorPair r (c.find (m.name ()));

          for (; r.first != r.second; ++r.first)
          {
            if (r.first->named ().is_a<Type> ())
            {
              Type& bm (dynamic_cast<Type&> (r.first->named ()));

              if (typeid (bm) != typeid (m))
                continue;

              if (m.qualified ())
              {
                if (bm.qualified () &&
                    m.name () == bm.name () &&
                    m.namespace_ ().name () == bm.namespace_ ().name ())
                  return bm;
              }
              else
              {
                if (!bm.qualified () && m.name () == bm.name ())
                  return bm;
              }
            }
          }

          // If we didn't find anything, try our base.
          //
          if (c.inherits_p ())
          {
            SemanticGraph::Type& base (c.inherits ().base ());

            if (base.is_a<Complex> ())
              return find_member (dynamic_cast<Complex&> (base), m);
          }

          //std::wcerr << "unable to find member " << m.name () << " in "
          //           << c.name () << std::endl;

          throw NotFound ();
        }

      private:
        NameSet& set_;
        SemanticGraph::Complex& base_;
      };


      //
      //
      struct Complex: Traversal::Complex
      {
        Complex (String const& name_ = "")
            : name (name_)
        {
        }

        virtual Void
        traverse (Type& c)
        {
          if (c.named ())
            name = c.name ();

          NameSet member_set;
          member_set.insert (name);

          // Add our base's members to the initial list unless we are
          // inheriting by restriction in which case we need to have
          // the same names as our base.
          //
          Boolean restriction (false);

          if (c.inherits_p ())
          {
            // @@ What if this types name is the same as one of base's
            //    members?
            //
            SemanticGraph::Type& base (c.inherits ().base ());

            if (base.is_a<SemanticGraph::Complex> ())
            {
              if (!base.context ().count (
                    "cxx-parser-name-processor-member-set"))
              {
                Complex t; // Base type is always named.
                t.dispatch (base);
              }

              NameSet const& base_set (
                base.context ().get<NameSet> (
                  "cxx-parser-name-processor-member-set"));

              member_set.insert (base_set.begin (), base_set.end ());

              // Inheritance by restriction from anyType is a special case.
	      //
              restriction = c.inherits ().is_a<SemanticGraph::Restricts> () &&
	        !c.inherits ().base ().is_a<SemanticGraph::AnyType> ();
            }
          }

          if (restriction)
          {
            MemberInRestriction member (
              member_set,
              dynamic_cast<SemanticGraph::Complex&> (
                c.inherits ().base ()));

            Traversal::Names names (member);

            Traversal::Complex::names (c, names);
          }
          else
          {
            Member member (member_set);
            Traversal::Names names (member);

            Traversal::Complex::names (c, names);
          }

          // @@ I am not removing this at the moment.
          //
          c.context ().set ("cxx-parser-name-processor-member-set",
                            member_set);
        }

      private:
        String name;
      };


      //
      //
      struct GlobalType: Traversal::Type
      {
        GlobalType (NameSet& set)
            : set_ (set)
        {
        }

        virtual Void
        traverse (SemanticGraph::Type& t)
        {
          find_name (t, set_);
        }

      private:
        NameSet& set_;
      };


      struct Namespace: Traversal::Namespace, Context
      {
        Namespace (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& ns)
        {
          NameSet& type_set (global_type_names[ns.name ()]);

          GlobalType type (type_set);
          Traversal::Names names (type);

          Context::ns = &ns;

          Traversal::Namespace::names (ns, names);
          Traversal::Namespace::names (ns);

          Context::ns = 0;
        }
      };


      // Go into implied/included/imported schemas while making sure
      // we don't process the same stuff more than once.
      //
      struct Uses: Traversal::Uses
      {
        virtual Void
        traverse (Type& u)
        {
          SemanticGraph::Schema& s (u.schema ());

          if (!s.context ().count ("cxx-parser-name-processor-seen"))
          {
            s.context ().set ("cxx-parser-name-processor-seen", true);
            Traversal::Uses::traverse (u);
          }
        }
      };

      Void
      process_impl (SemanticGraph::Schema& tu,
                    SemanticGraph::Path const& file)
      {
        Context ctx (tu, file);

        Traversal::Schema schema;
        Uses uses;

        schema >> uses >> schema;

        Traversal::Names schema_names;
        Namespace ns (ctx);
        Traversal::Names ns_names;

        schema >> schema_names >> ns >> ns_names;

        Complex complex;

        ns_names >> complex;

        // Some twisted schemas do recusive inclusions.
        //
        tu.context ().set ("cxx-parser-name-processor-seen", true);

        schema.dispatch (tu);
      }
    }

    Void NameProcessor::
    process (SemanticGraph::Schema& tu,
             SemanticGraph::Path const& file)
    {
      process_impl (tu, file);
    }
  }
}
